/**************************************************
Mex-C code for reproducing experiment 2: find and sketch over multiple scales
***************************************************/
# include <stdio.h>
# include <stdlib.h>
# include "mex.h"        /* the algorithm is connect to matlab */
# include "math.h"
# define PI 3.1415926
# define ABS(x) ((x)>0? (x):(-(x)))
# define MAX(x, y) ((x)>(y)? (x):(y))
# define MIN(x, y) ((x)<(y)? (x):(y))

double *double_vector(int n)
{
    double *v; 
    
    v = (double*) calloc (n, sizeof(double));
    
    return v; 
}

int *int_vector(int n)
{
    int *v; 
    
    v = (int*) calloc (n, sizeof(int));
    
    return v; 
}

double **double_matrix(int m, int n)
{
    double **mat; 
    int i; 
    
    mat = (double**) calloc(m, sizeof(double*)); 
    for (i=0; i<m; i++)
        mat[i] = double_vector(n); 
    
    return mat; 
}

int **int_matrix(int m, int n)
{
    int **mat; 
    int i; 
    
    mat = (int**) calloc(m, sizeof(int*)); 
    for (i=0; i<m; i++)
        mat[i] = int_vector(n); 
    
    return mat; 
}


/* getting the index of matlab image, (x,y) location, (sx, sy) sizes */
int px(int x, int y, int bx, int by)    
{            
   return (x + (y-1)*bx - 1); 
 }
 
 /* variables */
 int n;                      /* number of images */
 int N;                      /* number of orientations */
 double **fI;                /* filtered images in bounding box */
 double **fJ;                /* filtered images */
 double **mJ;                /* local maximum pooling for filtered images */
 double **C;                 /* inhibition coefficients */
 int h;                      /* halfsizes of filters */
 int sx0, sy0;                /* current size of image */
 int ax, ay;                 /* half size of bounding box of learned model */ 
 double T;                   /* threshold of selected bases */
 double *Mi, *Mx, *My, *Mm;  /* storing selected bases */
 double **allsymbol;         /* symbol of filters */
 int L, ore;                 /* allowed ranges of shifting in location and orientation */
 int sub;                    /* subsampling */
 double **tss;                /* map of log-likelihood score */
 double Upperbound;          /* satuation level for response */
 int Totalsketch;            /* total number of bases */
 double *sym;                 /* learned template */
 double **Asym;               /* symbolic sketch */
 double *Sx, *Sy;           /* sizes of images */
 int nstep;                 /* number of iterations */
 int xhere, yhere, ihere;   /* given bounding box on one example */
 int switchdraw; 
 int LIK;   /* If LIK = 1, we use log-likelihood score. If LIK = 0, we use active correlation */
 int **sinsh, **cossh;       /* store the shift to avoid repeated sin and cosin computation */

 
 /* store the shift values so that we do not need to repeat the sin and cos computation */ 
void storeshift()
{
    int ind, l;
    double theta; 
    
    sinsh = int_matrix(N, L+L+1); 
    cossh = int_matrix(N, L+L+1); 
    for (ind=0; ind<N; ind++)        
    {
        theta = PI*ind/N; 
        for (l=-L; l<=L; l++)
         {
            sinsh[ind][l+L] = floor(l*sub*sin(theta)+.5); 
            cossh[ind][l+L] = floor(l*sub*cos(theta)+.5); 
        }   
    }
}
/* for Gabor(x, y, orientation = ind), find local maximum in image i */
double shiftmax(double **fw, int i, int ind, int x, int y, int *rm, int sx, int sy) 
{
   double m;
   int x1, y1, l, ml, mx, my, de, d, d1, here, mo, md; 
 
   m = 0.; 
   for (l=-L; l<=L; l++)   
     {
        x1 = x + cossh[ind][l+L]; 
        y1 = y + sinsh[ind][l+L];   /* shifting in normal direction */
        if ((x1>=1)&&(x1<=sx)&&(y1>=1)&&(y1<=sy))
        {
        here = px(x1, y1, sx, sy);
        for (de=-ore; de<=ore; de++)
            {
              d = de+ind;
              d1 = d;
              if (d<0)
                 d1 = d+N;
              if (d>=N)
                  d1 = d-N;   /* shifting in orientation */
              if (fw[d1*n+i][here]>m)
                {
                   m = fw[d1*n+i][here];   /* local maximization */
                   ml = l; mx = x1; my = y1; mo = d1; md = de; 
                }
             }
        }
     }
   rm[0] = ml; rm[1] = mx; rm[2] = my; rm[3] = mo; rm[4] = md; 
   return(MIN(Upperbound, m)); 
}

/* symbolc sketch */
void draw(double *sym, int x0, int y0, int ind, double w, int sx, int sy)
{
  int x, y; 
  double a; 
          
  for (x=x0-h; x<=x0+h; x++)
     for (y=y0-h; y<=y0+h; y++)
       {
         a = allsymbol[ind][px(x-x0+h+1, y-y0+h+1, 2*h+1, 2*h+1)]*w; 
         if (sym[px(x, y, sx, sy)]<a)
             sym[px(x, y, sx, sy)] = a;
       }
}

/* inhibit nearby Gabors */
void inhibit(int i, int mi, int mx, int my) 
{
   int x0, y0, ind; 
   double *f, *fc; 
   
   for (ind=0; ind<N; ind++)   
     {
       f = fI[ind*n+i];   
       fc = C[mi+ind*N];   /* inhibition coefficients, zero = inhibit */
       for (x0=MAX(1, mx-2*h); x0<=MIN(sx0, mx+2*h); x0++)
         for (y0=MAX(1, my-2*h); y0<=MIN(sy0, my+2*h); y0++)
         {
          f[px(x0, y0, sx0, sy0)] *= fc[px(x0-mx+2*h+1, y0-my+2*h+1, 4*h+1, 4*h+1)];
         }
      }
}

/* local maximization */
void Cmm(int i)
{
     int  ind, x, y, rm[5], sx, sy;  

     sx = floor(Sx[i]+.5); sy = floor(Sy[i]+.5);
     for (ind=0; ind<N; ind++)
     { 
        for (x=floor((h+1)/sub)+1+L; x<=floor((sx-h)/sub)-1-L; x++)
           for (y=floor((h+1)/sub)+1+L; y<=floor((sy-h)/sub)-1-L; y++)
             {
                   mJ[ind*n+i][px(x*sub, y*sub, sx, sy)] = shiftmax(fJ, i, ind, x*sub, y*sub, rm, sx, sy);   /* local maximum of active Gabor */  
               }
      }
}

/* copy the annotated bounding box of J image into I image */
void Ccopy()
{
    int i, ind, x, y, sx, sy;
    
    i = ihere; 
    sx = floor(Sx[i]+.5); sy = floor(Sy[i]+.5);
    for (ind=0; ind<N; ind++)
     { 
        for (x=1; x<=sx0; x++)
           for (y=1; y<=sy0; y++)
           {
               fI[ind*n+i][px(x, y, sx0, sy0)] = fJ[ind*n+i][px(xhere+x, yhere+y, sx, sy)]; 
           }
     }
}

/* fit the I image within the annotated bounding box */
void Cmp0()
{
   int i, ind, x, y, mi, mx, my, t; 
   double m, s; 

   i = ihere; 
   t = 0; 
   
   Ccopy();      /* copy from annotated image */
   /* fit the single image within the given bounding box */
   do  
   {
     m = 0.; 
     for (ind=0; ind<N; ind++)
     { 
        for (x=floor((h+1)/sub)+1+L; x<=floor((sx0-h)/sub)-1-L; x++)
           for (y=floor((h+1)/sub)+1+L; y<=floor((sy0-h)/sub)-1-L; y++)
             {
               s = fI[ind*n+i][px(x*sub, y*sub, sx0, sy0)];    
               if (m<s)
                 {
                    m = s; 
                    mi = ind; mx = x*sub; my = y*sub; 
                 }
              }
       }
     Mi[t] = mi; Mx[t] = mx; My[t] = my; 
     if (LIK)
     {
       Mm[t] = m;   
     }
     else
     {
        Mm[t] = sqrt(m); 
     }
     inhibit(i, mi, mx, my); 
         
     t++; 
   }
  while (t<Totalsketch);   
}


/* fit the I-images within the detected bounding boxes */
/* the same code as in experiment 1 */
void Cmp()
{
   int i, ind, x, y, mi, mx, my, t, rm[5]; 
   double m, s, mm; 

   t = 0; 
   do  
   {
     m = 0.; 
     for (ind=0; ind<N; ind++)
     { 
        for (x=floor((h+1)/sub)+1+L; x<=floor((sx0-h)/sub)-1-L; x++)
           for (y=floor((h+1)/sub)+1+L; y<=floor((sy0-h)/sub)-1-L; y++)
             {
               s = 0.; 
               for (i=0; i<n; i++)
                 {
                   mm = shiftmax(fI, i, ind, x*sub, y*sub, rm, sx0, sy0);   /* local maximum of active Gabor */
                   if (LIK)
                   {
                    s += mm; 
                   }
                   else
                   {
                   s += sqrt(mm); 
                   }
                  }   
                if (m<s)
                  {
                    m = s; 
                    mi = ind; mx = x*sub; my = y*sub; 
                   }
                }
         }
     Mi[t] = mi; Mx[t] = mx; My[t] = my; 
     Mm[t] = m/n;   

     if (switchdraw)
         draw(sym, mx, my, mi, Mm[t], sx0, sy0);       
           
     for (i=0; i<n; i++)
        {            
          mm = shiftmax(fI, i, mi, mx, my, rm, sx0, sy0);    
          if (mm>0.)
            {  
               inhibit(i, rm[3], rm[1], rm[2]); 
            }
        }
      t++; 
   }
  while (t<Totalsketch);   
}

/* find the bounding boxes of the J-images using the currently fitted model */
/* the same as the second experiment, here we do not do scale selection */
void Cfind(int i)
{
   int x0, y0, mi, mx, my, t, Fx, Fy, rm[5], x, y, ind, sx, sy; 
   double F, mm; 

   sx = floor(Sx[i]+.5); sy = floor(Sy[i]+.5);
   t = 0; 
   do  
   {
     mi = Mi[t]; mx = Mx[t]; my = My[t];  
     /* the following is a parallel search */
        
     for (x=1; x<=floor((sx-ax*2)/sub); x++)
          for (y=1; y<=floor((sy-ay*2)/sub); y++)
           {
             x0 = x*sub; y0 = y*sub; 
             if (LIK)
             {
              tss[i][px(x0+ax, y0+ay, sx, sy)] += (1. - 1./Mm[t])*mJ[mi*n+i][px(x0+mx, y0+my, sx, sy)] - log(Mm[t]); 
             }
           else
             {
             tss[i][px(x0+ax, y0+ay, sx, sy)] += Mm[t]*sqrt(mJ[mi*n+i][px(x0+mx, y0+my, sx, sy)]); 
             }
           }   
     t++; 
     }
  while (t<Totalsketch);   /* stopping criterion */
   
  /* find the center pixel of the best score */ 
  F = 0.; Fx = 0; Fy = 0;
  for (x=1; x<=floor((sx-ax*2)/sub); x++)
          for (y=1; y<=floor((sy-ay*2)/sub); y++)
           {
             x0 = x*sub; y0 = y*sub;      
             if (F<tss[i][px(x0+ax, y0+ay, sx, sy)]) 
             {
                 F = tss[i][px(x0+ax, y0+ay, sx, sy)];
                 Fx = x0; 
                 Fy = y0; 
             }
         }
 
  /* copy the J-images within the detected bounding boxes into I-images */
  for (ind=0; ind<N; ind++)
     {        
        for (x=1; x<=sx0; x++)
           for (y=1; y<=sy0; y++)
           {
               fI[ind*n+i][px(x, y, sx0, sy0)] = fJ[ind*n+i][px(Fx+x, Fy+y, sx, sy)]; 
           }
     }
  
 /* at the end of algorithm, sketch the detected objects */
 if(switchdraw)
  {
   t = 0; 
   do  
   {
     mi = Mi[t]; mx = Mx[t]; my = My[t];   
     mm = shiftmax(fJ, i, mi, Fx+mx, Fy+my, rm, sx, sy);  
     draw(Asym[i], rm[1], rm[2], rm[3], sqrt(mm), sx, sy); 
     t++; 
     }
  while (t<Totalsketch);   /* stopping criterion */ 
  }
}


/* mex function is used to pass on the pointers and scalars from matlab, 
   so that heavy computation can be done by C, which puts the results into 
   some of the pointers. After that, matlab can then use these results. 
   
   So matlab is very much like a managing platform for organizing the 
   experiments, and mex C is like a work enginee for fast computation. */

void mexFunction(int nlhs, mxArray *plhs[], 
                 int nrhs, const mxArray *prhs[])                
{
 int i, ind, j, c; 
 mxArray *f;  
 
 c = 0; /* counter for input pointers and scalars */
 n = floor(mxGetScalar(prhs[c++])+.5);
 N = floor(mxGetScalar(prhs[c++])+.5);  /* number of orientations */
  
 fI = mxCalloc(n*N, sizeof(double*));   /* fitered images */
 for (i=0; i<n; i++)
   {
     for (ind=0; ind<N; ind++)
      {  
       f = mxGetCell(prhs[c], ind*n+i); 
       fI[ind*n+i] = mxGetPr(f);    /* get pointers to filtered images */       
      }
    }
 c++;
 
 fJ = mxCalloc(n*N, sizeof(double*));   /* fitered images */
 for (i=0; i<n; i++)
   {
      for (ind=0; ind<N; ind++)
      {  
       f = mxGetCell(prhs[c], ind*n+i); 
       fJ[ind*n+i] = mxGetPr(f);    /* get pointers to filtered images */       
      }
    }
 c++;

 mJ = mxCalloc(n*N, sizeof(double*));   /* fitered images */
 for (i=0; i<n; i++)
   {
      for (ind=0; ind<N; ind++)
      {  
       f = mxGetCell(prhs[c], ind*n+i); 
       mJ[ind*n+i] = mxGetPr(f);    /* get pointers to filtered images */       
      }
    }
 c++;
 
 C = mxCalloc(N*N, sizeof(double*));    /* C: correlation/inhibition between filters */
 for (ind=0; ind<N; ind++)
     {  
       for (j=0; j<N; j++)
        {
         f = mxGetCell(prhs[c], j*N+ind); 
         C[j*N+ind] = mxGetPr(f);         /* get correlation/inhibition */
        }   
     }
 c++; 
 
 h = floor(mxGetScalar(prhs[c++])+.5);    /* half size of filters */
 
 sx0 = floor(mxGetScalar(prhs[c++])+.5);
 sy0 = floor(mxGetScalar(prhs[c++])+.5);   /* size of window */
 ax = floor(sx0/2. + .5); ay = floor(sy0/2. + .5); 
 
 T = mxGetScalar(prhs[c++]);              /* threshold */  
     
 Mi = mxGetPr(prhs[c++]);   /* orientation of selected Gabor */               
 Mx = mxGetPr(prhs[c++]);         
 My = mxGetPr(prhs[c++]);   /* (x, y) position of selected Gabor */
 Mm = mxGetPr(prhs[c++]);   /* average response of selected Gabor */ 
 
 allsymbol = mxCalloc(N, sizeof(double*));    
 for (ind=0; ind<N; ind++)
     {  
       f = mxGetCell(prhs[c], ind); 
       allsymbol[ind] = mxGetPr(f);  /* symbols of filters */         
     }
 c++; 
 
 L = floor(mxGetScalar(prhs[c++])+.5);     /* range of shifting along normal direction of Gabor */
 ore = floor(mxGetScalar(prhs[c++])+.5);   /* range of shifting in orientation */
 sub = floor(mxGetScalar(prhs[c++])+.5);   /* sub-sampling of Gabor filters to improve speed */
 Totalsketch = floor(mxGetScalar(prhs[c++])+.5);  /* total number of bases */
 Upperbound = mxGetScalar(prhs[c++]);             /* satuation level */
   
 
  tss = mxCalloc(n, sizeof(double*)); 
  for (i=0; i<n; i++)
   {
         f = mxGetCell(prhs[c], i); 
         tss[i] = mxGetPr(f);
   } 
 c++; 
/* map of log-likelihood score */
 
 sym = mxGetPr(prhs[c++]);   /* symbols of selected Gabors for all the images */       
 Asym = mxCalloc(n, sizeof(double*)); 
 for (i=0; i<n; i++)
   {
     f = mxGetCell(prhs[c], i); 
     Asym[i] = mxGetPr(f);  
  } 
 c++; 
 
 Sx = mxGetPr(prhs[c++]);
 Sy = mxGetPr(prhs[c++]);   /* sizes of J-images */
 
 nstep = floor(mxGetScalar(prhs[c++])+.5);   /* number of iterations */
 xhere = floor(mxGetScalar(prhs[c++])+.5); 
 yhere = floor(mxGetScalar(prhs[c++])+.5);   /* position of the given bounding box */
 ihere = floor(mxGetScalar(prhs[c++])+.5);   /* the image annoted by bounding box */

 LIK = floor(mxGetScalar(prhs[c++])+.5);    /* log-likelihood or active correlation? */
 
 storeshift();  
  
 switchdraw = 0;

 for (i=0; i<n; i++)
 {
 Cmm(i);      /* parallel local maximum pooling */  
 }
 
 Cmp0();  /* fit the given single bounding box */

 for (j=0; j<nstep; j++)
 {   
     if (j==nstep-1)
         switchdraw = 1;  
     for (i=0; i<n; i++)
	 {
            Cfind(i);   /* use the current model to find bounding boxes */
	 }      
	 Cmp();  /* fit the images within the detected bounding boxes */ 
 }
}

     



 

                    